/*
 * Created on Jan 17, 2010
 * Created by Paul Gardner
 * 
 * Copyright 2010 Vuze, Inc.  All rights reserved.
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License only.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 */


package com.aelitis.azureus.plugins.antithrottle;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.gudy.azureus2.core3.util.DisplayFormatters;
import org.gudy.azureus2.plugins.Plugin;
import org.gudy.azureus2.plugins.PluginException;
import org.gudy.azureus2.plugins.PluginInterface;
import org.gudy.azureus2.plugins.download.Download;
import org.gudy.azureus2.plugins.download.DownloadManagerStats;
import org.gudy.azureus2.plugins.download.DownloadStats;
import org.gudy.azureus2.plugins.logging.LoggerChannel;
import org.gudy.azureus2.plugins.logging.LoggerChannelListener;
import org.gudy.azureus2.plugins.peers.PeerManager;
import org.gudy.azureus2.plugins.torrent.TorrentAttribute;
import org.gudy.azureus2.plugins.ui.UIManager;
import org.gudy.azureus2.plugins.ui.config.BooleanParameter;
import org.gudy.azureus2.plugins.ui.config.IntParameter;
import org.gudy.azureus2.plugins.ui.config.Parameter;
import org.gudy.azureus2.plugins.ui.model.BasicPluginConfigModel;
import org.gudy.azureus2.plugins.ui.model.BasicPluginViewModel;
import org.gudy.azureus2.plugins.ui.tables.TableCell;
import org.gudy.azureus2.plugins.ui.tables.TableCellRefreshListener;
import org.gudy.azureus2.plugins.ui.tables.TableColumn;
import org.gudy.azureus2.plugins.ui.tables.TableManager;
import org.gudy.azureus2.plugins.utils.LocaleUtilities;
import org.gudy.azureus2.plugins.utils.UTTimerEvent;
import org.gudy.azureus2.plugins.utils.UTTimerEventPerformer;


public class 
AntiThrottlePlugin 
	implements Plugin
{
	private PluginInterface		plugin_interface;
	private LoggerChannel		logger;
	
	public void 
	initialize(
		PluginInterface _plugin_interface )
	
		throws PluginException 
	{
		plugin_interface	= _plugin_interface;
		
		logger				= plugin_interface.getLogger().getTimeStampedChannel( "AntiThrottle" ); 

		final TorrentAttribute atp_ta = plugin_interface.getTorrentManager().getPluginAttribute( "atp_state" );
		
		LocaleUtilities loc_utils = plugin_interface.getUtilities().getLocaleUtilities();

		loc_utils.integrateLocalisedMessageBundle( "com.aelitis.azureus.plugins.antithrottle.internat.Messages" );

		UIManager	ui_manager	= plugin_interface.getUIManager();
		
		final BasicPluginViewModel	view_model = ui_manager.createBasicPluginViewModel( "azantithrottle.name" );

		view_model.getActivity().setVisible( false );
		view_model.getProgress().setVisible( false );
		
		logger.addListener(
				new LoggerChannelListener()
				{
					public void
					messageLogged(
						int		type,
						String	content )
					{
						view_model.getLogArea().appendText( content + "\n" );
					}
					
					public void
					messageLogged(
						String		str,
						Throwable	error )
					{
						if ( str.length() > 0 ){
							view_model.getLogArea().appendText( str + "\n" );
						}
						
						StringWriter sw = new StringWriter();
						
						PrintWriter	pw = new PrintWriter( sw );
						
						error.printStackTrace( pw );
						
						pw.flush();
						
						view_model.getLogArea().appendText( sw.toString() + "\n" );
					}
				});		
		
		BasicPluginConfigModel config_model = 
			ui_manager.createBasicPluginConfigModel( "azantithrottle.name" );

		view_model.setConfigSectionID( "azantithrottle.name" );
		
		final BooleanParameter 	ass_enable		= config_model.addBooleanParameter2( "azantithrottle.ass.enable", "azantithrottle.ass.enable", true );
		final IntParameter  	ass_min_speed 	= config_model.addIntParameter2( "azantithrottle.ass.min_down_speed", "azantithrottle.ass.min_down_speed", 25 );
		final IntParameter  	ass_trigger 	= config_model.addIntParameter2( "azantithrottle.ass.trigger_time", "azantithrottle.ass.trigger_time", 120 );
		final IntParameter  	ass_delay 		= config_model.addIntParameter2( "azantithrottle.ass.time_before_restart", "azantithrottle.ass.time_before_restart", 15 );
		final BooleanParameter 	ass_no_restart	= config_model.addBooleanParameter2( "azantithrottle.ass.norestart", "azantithrottle.ass.norestart", false );
		final BooleanParameter 	ass_ignore_forced	= config_model.addBooleanParameter2( "azantithrottle.ass.ignore_forced", "azantithrottle.ass.ignore_forced", false );
		final BooleanParameter 	ass_ignore_upping	= config_model.addBooleanParameter2( "azantithrottle.ass.ignore_upping", "azantithrottle.ass.ignore_upping", false );
		
		ass_enable.addEnabledOnSelection( ass_min_speed );
		ass_enable.addEnabledOnSelection( ass_trigger );
		ass_enable.addEnabledOnSelection( ass_delay );
		ass_enable.addEnabledOnSelection( ass_no_restart );
		ass_enable.addEnabledOnSelection( ass_ignore_forced );
		ass_enable.addEnabledOnSelection( ass_ignore_upping );
		
		config_model.createGroup(
			"azantithrottle.ass.name",
			new Parameter[]{ 
					ass_enable, ass_min_speed, ass_trigger, ass_delay, ass_no_restart,
					ass_ignore_forced, ass_ignore_upping });
		
				
		TableManager	table_manager = ui_manager.getTableManager();
		
		final String restart_text = loc_utils.getLocalisedMessageText( "azantithrottle.ass.restarting" );
		
			// status column
		
		for ( String col_name: new String[]{ TableManager.TABLE_MYTORRENTS_ALL_BIG, TableManager.TABLE_MYTORRENTS_INCOMPLETE, TableManager.TABLE_MYTORRENTS_INCOMPLETE_BIG }){
			
			TableColumn	status_column = 
				table_manager.createColumn( col_name, "azantithrottle.ass.column.name" );
			
			status_column.setAlignment(TableColumn.ALIGN_LEAD);
			status_column.setPosition( TableColumn.POSITION_LAST );
			status_column.setRefreshInterval(TableColumn.INTERVAL_LIVE);
			status_column.setType(TableColumn.TYPE_TEXT);
			status_column.setWidth( 100 );
			
			status_column.setVisible( true );
			
			status_column.addCellRefreshListener(
				new TableCellRefreshListener()
				{
					public void 
					refresh(
						TableCell cell )
					{
						Download	download = (Download)cell.getDataSource();
						
						if ( download != null ){
							
							String	text;
						
							String str = download.getAttribute( atp_ta );

							if ( str != null && str.length() > 0 ){
								
								text = restart_text;
							}else{
								
								text = "";
							}
						
							cell.setText( text );
						}
					}
				});
			
			table_manager.addColumn( status_column );
		}
		
		logger.log( "Enabled=" + ass_enable.getValue());
		
		plugin_interface.getUtilities().createTimer( "processor", true ).addPeriodicEvent(
			1000,
			new UTTimerEventPerformer()
			{
				private int	tick_count		= 0;
				
				private static final int PS_MONITORING		= 0;
				private static final int PS_STOPPING_1		= 1;
				private static final int PS_STOPPING_2		= 2;
				private static final int PS_RESTART_WAIT	= 3;
				private static final int PS_RESTARTING_1	= 4;
				private static final int PS_RESTARTING_2	= 5;
				
				private int	process_state	= PS_MONITORING;
				private int trigger_secs 	= 0;
				private int	wait_secs		= 0;
				
				private List<Download>	stopped_downloads  	= new ArrayList<Download>();
				private List<Download>	held_downloads  	= new ArrayList<Download>();
				
				public void 
				perform(
					UTTimerEvent event )
				{
					if ( tick_count == 0 ){
						
						Download[] downloads = plugin_interface.getDownloadManager().getDownloads();
						
						for ( Download download: downloads ){
						
							if ( download.getState() == Download.ST_STOPPED ){
											
								String str = download.getAttribute( atp_ta );
								
								if ( str != null && str.length() > 0 ){
								
									try{
										download.restart();
										
									}catch( Throwable e ){
										
										logger.log( "Failed to restart " + download.getName(), e );
									}
									
									if ( str.equals( "f" )){
									
										download.setForceStart( true );
									}
									
									download.setAttribute( atp_ta, "" );
								}
							}
						}
					}
					
					tick_count++;

					if ( !ass_enable.getValue()){
						
						process_state		= PS_MONITORING;
						trigger_secs 		= 0;
						wait_secs			= 0;
						
						stopped_downloads.clear();
						held_downloads.clear();
						
						return;
					}
										
					if ( tick_count % 30 == 0 ){
						
						for ( Iterator<Download> it=stopped_downloads.iterator();it.hasNext();){
	
							Download download = it.next();
							
							if ( download.isRemoved()){
								
								it.remove();
							}
						}
						
						for ( Iterator<Download> it=held_downloads.iterator();it.hasNext();){
	
							Download download = it.next();
							
							if ( download.isRemoved()){
								
								it.remove();
								
							}else if ( download.getLastScrapeResult().getSeedCount() > 0 ){
								
								it.remove();
								
								logger.log( "Download '" + download.getName() + "' has seeds - restarting" );
								
								try{
									download.restart();
									
								}catch( Throwable e ){
									
									logger.log( "Failed to restart '" + download.getName() + "'", e );
								}
							}
						}
					}
					
					if ( process_state == PS_MONITORING ){
						
							// track download speed
						
						DownloadManagerStats stats = plugin_interface.getDownloadManager().getStats();
						
						int	download_speed = stats.getDataAndProtocolReceiveRate();
						
						if ((download_speed/1024) <= ass_min_speed.getValue()){
							
							trigger_secs++;
							
							if ( trigger_secs >= ass_trigger.getValue()){
								
								trigger_secs = 0;
								
								logger.log( "Min download speed of " + ass_min_speed.getValue() + " KB/sec not reached for " + ass_trigger.getValue() + " secs - stopping downloads" );
								
								process_state = PS_STOPPING_1;
							}
						}else{
							
							trigger_secs = 0;
						}
					}else if ( process_state == PS_STOPPING_1 ){
						
							// stop the downloads
						
						Download[] downloads = plugin_interface.getDownloadManager().getDownloads();
						
						for ( Download download: downloads ){
							
							if ( download.getState() == Download.ST_DOWNLOADING && !download.isComplete( false ) ){
								
								if ( download.isForceStart() && ass_ignore_forced.getValue()){
									
									logger.log( "Not stopping '" + download.getName() + "' as force-started" );
									
									continue;
								}
								
								DownloadStats	stats = download.getStats();
								
								if ( stats.getUploadAverage() >= 1024 && ass_ignore_upping.getValue()){
									
									logger.log( "Not stopping '" + download.getName() + "' as currently uploading" );
									
									continue;
								}
								
								PeerManager pm = download.getPeerManager();
								
								if ( pm != null ){
									
									try{
										logger.log( 
											"    Stopping " + download.getName() + 
											": done=" + DisplayFormatters.formatPercentFromThousands( stats.getDownloadCompleted( false )) +
											", down=" + DisplayFormatters.formatByteCountToKiBEtcPerSec( stats.getDownloadAverage()) + 
											", up=" + DisplayFormatters.formatByteCountToKiBEtcPerSec( stats.getUploadAverage()) +
											", con_leechers=" + pm.getStats().getConnectedLeechers() +
											", con_seeds=" + pm.getStats().getConnectedSeeds());
											
										boolean was_force = download.isForceStart();
										
										download.setAttribute( atp_ta, was_force?"f":"n" );
										
										download.stop();
										
										stopped_downloads.add( download );
									
									}catch( Throwable e ){
									
										logger.log( "Failed to stop '" + download.getName() + "'", e );
									}
								}	
							}
						}
						
						if ( stopped_downloads.size() == 0 ){
						
							logger.log( "No downloads found to stop" );
							
						}else{
							
							logger.log( "Stopped " + stopped_downloads.size() + " downloads - waiting for " + ass_delay.getValue() + " secs before restarting" );
						}
						
						process_state = PS_STOPPING_2;
						
					}else if ( process_state == PS_STOPPING_2 ){
						
						Iterator<Download> it = stopped_downloads.iterator();
						
						boolean	wait = false;
						
						while( it.hasNext()){
							
							Download d = it.next();
							
							if ( d.isRemoved() || d.getState() == Download.ST_ERROR ){
								
								it.remove();
								
							}else{
								
								int	state = d.getState();
								
								if ( state == Download.ST_STOPPING ){
									
									wait	= true;
									
								}else if ( state == Download.ST_STOPPED ){
									
								}else{
									
									try{
										logger.log( "    Re-stopping " + d.getName());
										
										wait	= true;
										
										d.stop();
										
									}catch( Throwable e ){
										
									}
								}
							}
						}
						
						if ( wait ){
							
							logger.log( "    Waiting for downloads to stop..." );
							
						}else{
							
							process_state = PS_RESTART_WAIT;
						}
					}else if ( process_state == PS_RESTART_WAIT ){
						
							// wait for restart
						
						wait_secs++;
						
						if ( wait_secs >= ass_delay.getValue()){
							
							wait_secs = 0;
							
							logger.log( "Restarting downloads" );
							
							process_state = PS_RESTARTING_1;
						}
						
					}else if ( process_state == PS_RESTARTING_1 ){
						
						for ( Download download: stopped_downloads ){
							
							if ( ass_no_restart.getValue()){
								
								if ( download.getLastScrapeResult().getSeedCount() <= 0 ){
									
									if ( !held_downloads.contains( download )){
									
										held_downloads.add( download );
									
										logger.log( "Download '" + download.getName() + "' has no seeds - not restarting" );
									}
									
									continue;
								}
							}
							
							try{
								logger.log( "    Starting " + download.getName());
								
								download.restart();
								
								String str = download.getAttribute( atp_ta );
								
								if ( str != null && str.length() > 0 ){
									
									if ( str.equals( "f" )){
									
										download.setForceStart( true );
									}
									
									download.setAttribute( atp_ta, "" );
								}
							}catch( Throwable e ){
								
								logger.log( "Failed to restart '" + download.getName() + "'", e );
							}
						}
						
						logger.log( "Restarted " + stopped_downloads.size() + " downloads" );
												
							// retart
						
						process_state = PS_RESTARTING_2;
						
					}else{
						
						Iterator<Download> it = stopped_downloads.iterator();
						
						boolean	wait = false;
						
						while( it.hasNext()){
							
							Download d = it.next();
							
							if ( d.isRemoved() || d.getState() == Download.ST_ERROR ){
								
								it.remove();
								
							}else{
								
								int	state = d.getState();
								
								if ( state == Download.ST_STOPPED ){
									
									try{
										logger.log( "    Re-starting " + d.getName());
										
										wait = true;

										d.restart();
										
									}catch( Throwable e ){
										
									}									
								}
							}
						}
						
						if ( wait ){
						
							logger.log( "    Waiting for downloads to restart..." );
							
						}else{
							
							stopped_downloads.clear();

							process_state = PS_MONITORING;
						}
					}

				}
			});
	}
}
